//-----------------------------------------------------------------------------
// vflt_sse2.h : LPF(SSE2-Int)
// Copyright (c) 2005-2008  N099
//-----------------------------------------------------------------------------
#include "vflt.h"

// d
static const u128i_i16 xc0_3 = { 2, 2, 2, 2, 2, 2, 2, 2 };

static const u128i_i16 xc0_5 = { 8192, -2606, 8192, -2606, 8192, -2606, 8192, -2606 };
static const u128i_i16 xc1_5 = { 21596, 0, 21596, 0, 21596, 0, 21596, 0 };
static const u128i_i32 xc2_5 = { (1<<14)-1, (1<<14)-1, (1<<14)-1, (1<<14)-1 };

static const u128i_i16 xc0_7 = { -4000, 1315, -4000, 1315, -4000, 1315, -4000, 1315 };
static const u128i_i16 xc1_7 = {  6876, 24386, 6876, 24386, 6876, 24386, 6876, 24386 };
static const u128i_i32 xc2_7 = { (1<<14) - 1, (1<<14) - 1, (1<<14) - 1, (1<<14) - 1 };

static const u128i_i16 xc0_9 = { 2285, -766, 2285, -766, 2285, -766, 2285, -766 };
static const u128i_i16 xc1_9 = { 5907, -4237, 5907, -4237, 5907, -4237, 5907, -4237 };
static const u128i_i16 xc2_9 = { (1<<14)-1, 26390, (1<<14)-1, 26390, (1<<14)-1, 26390, (1<<14)-1, 26390 };
static const u128i_i16 xc3_9 = { 1, 1, 1, 1, 1, 1, 1, 1 };

static const u128i_i16 xc0_b = { -1442, 518, -1442, 518, -1442, 518, -1442, 518 };
static const u128i_i16 xc1_b = { -3993, 2693, -3993, 2693, -3993, 2693, -3993, 2693 };
static const u128i_i16 xc2_b = { 4980, 27256, 4980, 27256, 4980, 27256, 4980, 27256 };
static const u128i_i32 xc3_b = { (1<<14)-1, (1<<14)-1, (1<<14)-1, (1<<14)-1 };

static const u128i_i16 xc0_d = { 981, -363, 981, -363, 981, -363, 981, -363 };
static const u128i_i16 xc1_d = { 2822, -1841, 2822, -1841, 2822, -1841, 2822, -1841 };
static const u128i_i16 xc2_d = { 4389, -3737, 4389, -3737, 4389, -3737, 4389, -3737 };
static const u128i_i16 xc3_d = { (1<<14)-1, 28266, (1<<14)-1, 28266, (1<<14)-1, 28266, (1<<14)-1, 28266 };
static const u128i_i16 xc4_d = { 1, 1, 1, 1, 1, 1, 1, 1 };

static const u128i_i16 xc0_f = { -695, 265, -695, 265, -695, 265, -695, 265 };
static const u128i_i16 xc1_f = { -2021, 1299, -2021, 1299, -2021, 1299, -2021, 1299 };
static const u128i_i16 xc2_f = { -3417, 2766, -3417, 2766, -3417, 2766, -3417, 2766 };
static const u128i_i16 xc3_f = { 28650, 3862, 28650, 3862, 28650, 3862, 28650, 3862 };
static const u128i_i32 xc4_f = { (1<<14) - 1, (1<<14) - 1, (1<<14) - 1, (1<<14) - 1 };

// 3tap LPF(aligned)
NOA static void lpf_v_3_sse2(pix * restrict dst, const pix * restrict src,
                             int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2;
    p_m128r dst0;
    __m128i xs0,xs1,xs2;

    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src1 = (p_m128c)src;
        src0 = (p_m128c)(src + ((i-1) >= 0 ? -s : s));
        src2 = (p_m128c)(src + ((i+1) >= h ? -s : s));
        
        for(j=0; j<w; ++j){
            xs0 = _mm_load_si128(src0+j);
            xs1 = _mm_load_si128(src1+j);
            xs2 = _mm_load_si128(src2+j);

            xs0 = _mm_adds_epi16(xs0, xs2);
            xs1 = _mm_slli_epi16(xs1, 1);

            xs0 = _mm_adds_epi16(xs0, xc0_3.x);
            xs1 = _mm_adds_epi16(xs1, xs0);
            xs1 = _mm_srai_epi16(xs1, 2);
            
            _mm_store_si128(dst0+j, xs1);
        }
        
        dst += s;
        src += s;
  }
}

// 5tap LPF(aligned)
NOA static void lpf_v_5_sse2(pix * restrict dst, const pix * restrict src,
                             int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3;
    
    w = COUNT16(w * sizeof(pix));
  
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src2 = (p_m128c)src;
        src1 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src3 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src0 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src4 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_load_si128(src0+j), _mm_load_si128(src4+j));
            xs1 = _mm_adds_epi16(_mm_load_si128(src1+j), _mm_load_si128(src3+j));
            xs0 = xs1;
            
            xs1 = _mm_unpackhi_epi16(xs1, xs2);
            xs0 = _mm_unpacklo_epi16(xs0, xs2);
            
            xs1 = _mm_madd_epi16(xs1, xc0_5.x);
            xs0 = _mm_madd_epi16(xs0, xc0_5.x);
            
            xs0 = _mm_add_epi32(xs0, xc2_5.x);
            xs1 = _mm_add_epi32(xs1, xc2_5.x);

            xs2 = _mm_load_si128(src2+j);
            xs3 = xs2;
            xs3 = _mm_unpackhi_epi16(xs3, xc0_5.x);
            xs2 = _mm_unpacklo_epi16(xs2, xc0_5.x);
            
            xs2 = _mm_madd_epi16(xs2, xc1_5.x);
            xs3 = _mm_madd_epi16(xs3, xc1_5.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_store_si128(dst0+j, _mm_packs_epi32(xs0, xs1));
        }
        
        dst += s;
        src += s;
    }
}

// 7tap LPF(aligned)
NOA static void lpf_v_7_sse2(pix * restrict dst, const pix * restrict src,
                             int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5, src6;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;

    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src3 = (p_m128c)src;
        src2 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src4 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src1 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src5 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src0 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src6 = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_load_si128(src0+j), _mm_load_si128(src6+j)); 
            xs1 = _mm_adds_epi16(_mm_load_si128(src1+j), _mm_load_si128(src5+j));
            xs0 = xs1;
            
            xs1 = _mm_unpackhi_epi16(xs1, xs2);
            xs0 = _mm_unpacklo_epi16(xs0, xs2);
            
            xs0 = _mm_madd_epi16(xs0, xc0_7.x);
            xs1 = _mm_madd_epi16(xs1, xc0_7.x);
            
            xs0 = _mm_add_epi32(xs0, xc2_7.x);
            xs1 = _mm_add_epi32(xs1, xc2_7.x);
            
            xs4 = _mm_adds_epi16(_mm_load_si128(src2+j), _mm_load_si128(src4+j));
            xs3 = _mm_load_si128(src3+j);
            xs2 = xs4;
            xs4 = _mm_unpackhi_epi16(xs4, xs3);
            xs2 = _mm_unpacklo_epi16(xs2, xs3);
            
            xs4 = _mm_madd_epi16(xs4, xc1_7.x);
            xs2 = _mm_madd_epi16(xs2, xc1_7.x);
            
            xs1 = _mm_add_epi32(xs1, xs4);
            xs0 = _mm_add_epi32(xs0, xs2);
            
            xs1 = _mm_srai_epi32(xs1, 15);
            xs0 = _mm_srai_epi32(xs0, 15);
            
            _mm_store_si128(dst0+j, _mm_packs_epi32(xs0, xs1));
        }
        
        dst += s;
        src += s;
  }
}

// 9tap LPF(aligned)
NOA static void lpf_v_9_sse2(pix * restrict dst, const pix * restrict src,
                             int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5, src6, src7, src8;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;
    
    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src4 = (p_m128c)src;
        src3 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src5 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src2 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src6 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src1 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src7 = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src0 = (p_m128c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        src8 = (p_m128c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
    
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_load_si128(src0+j), _mm_load_si128(src8+j));
            xs1 = _mm_adds_epi16(_mm_load_si128(src1+j), _mm_load_si128(src7+j));
            xs0 = xs1;
            
            xs0 = _mm_unpackhi_epi16(xs0, xs2);
            xs1 = _mm_unpacklo_epi16(xs1, xs2);
            xs0 = _mm_madd_epi16(xs0, xc0_9.x);
            xs1 = _mm_madd_epi16(xs1, xc0_9.x);

            //
            xs4 = _mm_adds_epi16(_mm_load_si128(src2+j), _mm_load_si128(src6+j));
            xs3 = _mm_adds_epi16(_mm_load_si128(src3+j), _mm_load_si128(src5+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            xs2 = _mm_madd_epi16(xs2, xc1_9.x);
            xs3 = _mm_madd_epi16(xs3, xc1_9.x);
            
            //
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_load_si128(src4+j);
            xs2 = xs3 = xc3_9.x;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            xs2 = _mm_madd_epi16(xs2, xc2_9.x);
            xs3 = _mm_madd_epi16(xs3, xc2_9.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_store_si128(dst0+j, _mm_packs_epi32(xs1, xs0));
        }
        
        dst += s;
        src += s;
    }
}

// 11tap LPF(aligned)
NOA static void lpf_v_11_sse2(pix * restrict dst, const pix * restrict src,
                              int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5;
    p_m128c src6, src7, src8, src9, srca;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;
    
    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src5 = (p_m128c)src;
        src4 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src6 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src3 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src7 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src2 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src8 = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src1 = (p_m128c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        src9 = (p_m128c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src0 = (p_m128c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srca = (p_m128c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        
        for(j=0; j<w; ++j){
            //
            xs2 = _mm_adds_epi16(_mm_load_si128(src0+j), _mm_load_si128(srca+j));
            xs1 = _mm_adds_epi16(_mm_load_si128(src1+j), _mm_load_si128(src9+j));
            xs0 = xs1;
            
            xs0 = _mm_unpackhi_epi16(xs0, xs2);
            xs1 = _mm_unpacklo_epi16(xs1, xs2);
            xs0 = _mm_madd_epi16(xs0, xc0_b.x);
            xs1 = _mm_madd_epi16(xs1, xc0_b.x);
            
            xs0 = _mm_add_epi32(xs0, xc3_b.x);
            xs1 = _mm_add_epi32(xs1, xc3_b.x);
            
            //
            xs4 = _mm_adds_epi16(_mm_load_si128(src2+j), _mm_load_si128(src8+j));
            xs3 = _mm_adds_epi16(_mm_load_si128(src3+j), _mm_load_si128(src7+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            xs2 = _mm_madd_epi16(xs2, xc1_b.x);
            xs3 = _mm_madd_epi16(xs3, xc1_b.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs3 = _mm_adds_epi16(_mm_load_si128(src4+j), _mm_load_si128(src6+j));
            xs4 = _mm_load_si128(src5+j);
            xs2 = xs3;
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc2_b.x);
            xs3 = _mm_madd_epi16(xs3, xc2_b.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_store_si128(dst0+j, _mm_packs_epi32(xs1, xs0));
        }
        
        dst += s;
        src += s;
    }
}

// 13tap LPF(aligned)
NOA static void lpf_v_13_sse2(pix * restrict dst, const pix * restrict src,
                              int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5, src6;
    p_m128c src7, src8, src9, srca, srcb, srcc;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;
    
    w = COUNT16(w * sizeof(pix));

    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src6 = (p_m128c)src;
        src5 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src7 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src4 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src8 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src3 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src9 = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src2 = (p_m128c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        srca = (p_m128c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src1 = (p_m128c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srcb = (p_m128c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        src0 = (p_m128c)(src + s * ((i-6) >= 0 ? -6 : 6 - 2*i));
        srcc = (p_m128c)(src + s * ((i+6) >= h ? 2*(h-1-i)-6 : 6));
        
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_load_si128(src0+j), _mm_load_si128(srcc+j));
            xs1 = _mm_adds_epi16(_mm_load_si128(src1+j), _mm_load_si128(srcb+j));
            xs0 = xs1;
            
            xs0 = _mm_unpackhi_epi16(xs0, xs2);
            xs1 = _mm_unpacklo_epi16(xs1, xs2);
            
            xs0 = _mm_madd_epi16(xs0, xc0_d.x);
            xs1 = _mm_madd_epi16(xs1, xc0_d.x);
            
            //
            xs4 = _mm_adds_epi16(_mm_load_si128(src2+j), _mm_load_si128(srca+j));
            xs3 = _mm_adds_epi16(_mm_load_si128(src3+j), _mm_load_si128(src9+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc1_d.x);
            xs3 = _mm_madd_epi16(xs3, xc1_d.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_adds_epi16(_mm_load_si128(src4+j), _mm_load_si128(src8+j));
            xs3 = _mm_adds_epi16(_mm_load_si128(src5+j), _mm_load_si128(src7+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc2_d.x);
            xs3 = _mm_madd_epi16(xs3, xc2_d.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_load_si128(src6+j);
            xs3 = xs2 = xc4_d.x;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc3_d.x);
            xs3 = _mm_madd_epi16(xs3, xc3_d.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_store_si128(dst0+j, _mm_packs_epi32(xs1, xs0));
        }
        
        dst += s;
        src += s;
    }
}

// 15tap LPF(aligned)
NOA static void lpf_v_15_sse2(pix * restrict dst, const pix * restrict src,
                              int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5, src6, src7;
    p_m128c src8, src9, srca, srcb, srcc, srcd, srce;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;

    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src7 = (p_m128c)src;
        src6 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src8 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src5 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src9 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src4 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        srca = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src3 = (p_m128c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        srcb = (p_m128c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src2 = (p_m128c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srcc = (p_m128c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        src1 = (p_m128c)(src + s * ((i-6) >= 0 ? -6 : 6 - 2*i));
        srcd = (p_m128c)(src + s * ((i+6) >= h ? 2*(h-1-i)-6 : 6));
        src0 = (p_m128c)(src + s * ((i-7) >= 0 ? -7 : 7 - 2*i));
        srce = (p_m128c)(src + s * ((i+7) >= h ? 2*(h-1-i)-7 : 7));
        
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_load_si128(src0+j), _mm_load_si128(srce+j));
            xs1 = _mm_adds_epi16(_mm_load_si128(src1+j), _mm_load_si128(srcd+j));
            xs0 = xs1;
            
            xs0 = _mm_unpackhi_epi16(xs0, xs2);
            xs1 = _mm_unpacklo_epi16(xs1, xs2);
            
            xs0 = _mm_madd_epi16(xs0, xc0_f.x);
            xs1 = _mm_madd_epi16(xs1, xc0_f.x);
            
            xs0 = _mm_add_epi32(xs0, xc4_f.x);
            xs1 = _mm_add_epi32(xs1, xc4_f.x);
            
            //
            xs4 = _mm_adds_epi16(_mm_load_si128(src2+j), _mm_load_si128(srcc+j));
            xs3 = _mm_adds_epi16(_mm_load_si128(src3+j), _mm_load_si128(srcb+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);

            xs2 = _mm_madd_epi16(xs2, xc1_f.x);
            xs3 = _mm_madd_epi16(xs3, xc1_f.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_adds_epi16(_mm_load_si128(src4+j), _mm_load_si128(srca+j));
            xs3 = _mm_adds_epi16(_mm_load_si128(src5+j), _mm_load_si128(src9+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);

            xs2 = _mm_madd_epi16(xs2, xc2_f.x);
            xs3 = _mm_madd_epi16(xs3, xc2_f.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_adds_epi16(_mm_load_si128(src6+j), _mm_load_si128(src8+j));
            xs3 = _mm_load_si128(src7+j);
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc3_f.x);
            xs3 = _mm_madd_epi16(xs3, xc3_f.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_store_si128(dst0+j, _mm_packs_epi32(xs1, xs0));
        }
        
        dst += s;
        src += s;
    }
}


// 3tap LPF(not-aligned)
NOA static void lpf_v_3_sse2u(pix * restrict dst, const pix * restrict src,
                              int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2;
    p_m128r dst0;
    __m128i xs0,xs1,xs2;

    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src1 = (p_m128c)src;
        src0 = (p_m128c)(src + ((i-1) >= 0 ? -s : s));
        src2 = (p_m128c)(src + ((i+1) >= h ? -s : s));
        
        for(j=0; j<w; ++j){
            xs0 = _mm_loadu_si128(src0+j);
            xs1 = _mm_loadu_si128(src1+j);
            xs2 = _mm_loadu_si128(src2+j);

            xs0 = _mm_adds_epi16(xs0, xs2);
            xs1 = _mm_adds_epi16(xs1, xs1);

            xs0 = _mm_adds_epi16(xs0, xc0_3.x);
            xs1 = _mm_adds_epi16(xs1, xs0);
            xs1 = _mm_srai_epi16(xs1, 2);
            
            _mm_storeu_si128(dst0+j, xs1);
        }
        
        dst += s;
        src += s;
    }
}

// 5tap LPF(not-aligned)
NOA static void lpf_v_5_sse2u(pix * restrict dst, const pix * restrict src,
                              int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3;
    
    w = COUNT16(w * sizeof(pix));
  
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src2 = (p_m128c)src;
        src1 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src3 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src0 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src4 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_loadu_si128(src0+j), _mm_loadu_si128(src4+j));
            xs1 = _mm_adds_epi16(_mm_loadu_si128(src1+j), _mm_loadu_si128(src3+j));
            xs0 = xs1;
            
            xs1 = _mm_unpackhi_epi16(xs1, xs2);
            xs0 = _mm_unpacklo_epi16(xs0, xs2);
            
            xs1 = _mm_madd_epi16(xs1, xc0_5.x);
            xs0 = _mm_madd_epi16(xs0, xc0_5.x);
            
            xs0 = _mm_add_epi32(xs0, xc2_5.x);
            xs1 = _mm_add_epi32(xs1, xc2_5.x);

            xs2 = _mm_loadu_si128(src2+j);
            xs3 = xs2;
            xs3 = _mm_unpackhi_epi16(xs3, xc0_5.x);
            xs2 = _mm_unpacklo_epi16(xs2, xc0_5.x);
            
            xs2 = _mm_madd_epi16(xs2, xc1_5.x);
            xs3 = _mm_madd_epi16(xs3, xc1_5.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_storeu_si128(dst0+j, _mm_packs_epi32(xs0, xs1));
        }
        
        dst += s;
        src += s;
    }
}

// 7tap LPF(not-aligned)
NOA static void lpf_v_7_sse2u(pix * restrict dst, const pix * restrict src,
                              int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5, src6;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;

    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src3 = (p_m128c)src;
        src2 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src4 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src1 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src5 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src0 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src6 = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_loadu_si128(src0+j), _mm_loadu_si128(src6+j)); 
            xs1 = _mm_adds_epi16(_mm_loadu_si128(src1+j), _mm_loadu_si128(src5+j));
            xs0 = xs1;
            
            xs1 = _mm_unpackhi_epi16(xs1, xs2);
            xs0 = _mm_unpacklo_epi16(xs0, xs2);
            
            xs0 = _mm_madd_epi16(xs0, xc0_7.x);
            xs1 = _mm_madd_epi16(xs1, xc0_7.x);
            
            xs0 = _mm_add_epi32(xs0, xc2_7.x);
            xs1 = _mm_add_epi32(xs1, xc2_7.x);
            
            xs4 = _mm_adds_epi16(_mm_loadu_si128(src2+j), _mm_loadu_si128(src4+j));
            xs3 = _mm_loadu_si128(src3+j);
            xs2 = xs4;
            xs4 = _mm_unpackhi_epi16(xs4, xs3);
            xs2 = _mm_unpacklo_epi16(xs2, xs3);
            
            xs4 = _mm_madd_epi16(xs4, xc1_7.x);
            xs2 = _mm_madd_epi16(xs2, xc1_7.x);
            
            xs1 = _mm_add_epi32(xs1, xs4);
            xs0 = _mm_add_epi32(xs0, xs2);
            
            xs1 = _mm_srai_epi32(xs1, 15);
            xs0 = _mm_srai_epi32(xs0, 15);
            
            _mm_storeu_si128(dst0+j, _mm_packs_epi32(xs0, xs1));
        }
        
        dst += s;
        src += s;
    }
}
// 9tap LPF(not-aligned)
NOA static void lpf_v_9_sse2u(pix * restrict dst, const pix * restrict src,
                              int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5, src6, src7, src8;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;
    
    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src4 = (p_m128c)src;
        src3 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src5 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src2 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src6 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src1 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src7 = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src0 = (p_m128c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        src8 = (p_m128c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
    
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_loadu_si128(src0+j), _mm_loadu_si128(src8+j));
            xs1 = _mm_adds_epi16(_mm_loadu_si128(src1+j), _mm_loadu_si128(src7+j));
            xs0 = xs1;
            
            xs0 = _mm_unpackhi_epi16(xs0, xs2);
            xs1 = _mm_unpacklo_epi16(xs1, xs2);
            xs0 = _mm_madd_epi16(xs0, xc0_9.x);
            xs1 = _mm_madd_epi16(xs1, xc0_9.x);

            //
            xs4 = _mm_adds_epi16(_mm_loadu_si128(src2+j), _mm_loadu_si128(src6+j));
            xs3 = _mm_adds_epi16(_mm_loadu_si128(src3+j), _mm_loadu_si128(src5+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            xs2 = _mm_madd_epi16(xs2, xc1_9.x);
            xs3 = _mm_madd_epi16(xs3, xc1_9.x);
            
            //
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_loadu_si128(src4+j);
            xs2 = xs3 = xc3_9.x;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            xs2 = _mm_madd_epi16(xs2, xc2_9.x);
            xs3 = _mm_madd_epi16(xs3, xc2_9.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_storeu_si128(dst0+j, _mm_packs_epi32(xs1, xs0));
        }
        
        dst += s;
        src += s;
    }
}

// 11tap LPF(not-aligned)
NOA static void lpf_v_11_sse2u(pix * restrict dst, const pix * restrict src,
                               int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5;
    p_m128c src6, src7, src8, src9, srca;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;
    
    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src5 = (p_m128c)src;
        src4 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src6 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src3 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src7 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src2 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src8 = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src1 = (p_m128c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        src9 = (p_m128c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src0 = (p_m128c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srca = (p_m128c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        
        for(j=0; j<w; ++j){
            //
            xs2 = _mm_adds_epi16(_mm_loadu_si128(src0+j), _mm_loadu_si128(srca+j));
            xs1 = _mm_adds_epi16(_mm_loadu_si128(src1+j), _mm_loadu_si128(src9+j));
            xs0 = xs1;
            
            xs0 = _mm_unpackhi_epi16(xs0, xs2);
            xs1 = _mm_unpacklo_epi16(xs1, xs2);
            xs0 = _mm_madd_epi16(xs0, xc0_b.x);
            xs1 = _mm_madd_epi16(xs1, xc0_b.x);
            
            xs0 = _mm_add_epi32(xs0, xc3_b.x);
            xs1 = _mm_add_epi32(xs1, xc3_b.x);
            
            //
            xs4 = _mm_adds_epi16(_mm_loadu_si128(src2+j), _mm_loadu_si128(src8+j));
            xs3 = _mm_adds_epi16(_mm_loadu_si128(src3+j), _mm_loadu_si128(src7+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            xs2 = _mm_madd_epi16(xs2, xc1_b.x);
            xs3 = _mm_madd_epi16(xs3, xc1_b.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs3 = _mm_adds_epi16(_mm_loadu_si128(src4+j), _mm_loadu_si128(src6+j));
            xs4 = _mm_loadu_si128(src5+j);
            xs2 = xs3;
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc2_b.x);
            xs3 = _mm_madd_epi16(xs3, xc2_b.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_storeu_si128(dst0+j, _mm_packs_epi32(xs1, xs0));
        }
        
        dst += s;
        src += s;
    }
}

// 13tap LPF(not-aligned)
NOA static void lpf_v_13_sse2u(pix * restrict dst, const pix * restrict src,
                               int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5, src6;
    p_m128c src7, src8, src9, srca, srcb, srcc;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;
    
    w = COUNT16(w * sizeof(pix));

    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src6 = (p_m128c)src;
        src5 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src7 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src4 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src8 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src3 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src9 = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src2 = (p_m128c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        srca = (p_m128c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src1 = (p_m128c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srcb = (p_m128c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        src0 = (p_m128c)(src + s * ((i-6) >= 0 ? -6 : 6 - 2*i));
        srcc = (p_m128c)(src + s * ((i+6) >= h ? 2*(h-1-i)-6 : 6));
        
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_loadu_si128(src0+j), _mm_loadu_si128(srcc+j));
            xs1 = _mm_adds_epi16(_mm_loadu_si128(src1+j), _mm_loadu_si128(srcb+j));
            xs0 = xs1;
            
            xs0 = _mm_unpackhi_epi16(xs0, xs2);
            xs1 = _mm_unpacklo_epi16(xs1, xs2);
            
            xs0 = _mm_madd_epi16(xs0, xc0_d.x);
            xs1 = _mm_madd_epi16(xs1, xc0_d.x);
            
            //
            xs4 = _mm_adds_epi16(_mm_loadu_si128(src2+j), _mm_loadu_si128(srca+j));
            xs3 = _mm_adds_epi16(_mm_loadu_si128(src3+j), _mm_loadu_si128(src9+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc1_d.x);
            xs3 = _mm_madd_epi16(xs3, xc1_d.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_adds_epi16(_mm_loadu_si128(src4+j), _mm_loadu_si128(src8+j));
            xs3 = _mm_adds_epi16(_mm_loadu_si128(src5+j), _mm_loadu_si128(src7+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc2_d.x);
            xs3 = _mm_madd_epi16(xs3, xc2_d.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_loadu_si128(src6+j);
            xs3 = xs2 = xc4_d.x;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc3_d.x);
            xs3 = _mm_madd_epi16(xs3, xc3_d.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_storeu_si128(dst0+j, _mm_packs_epi32(xs1, xs0));
        }
        
        dst += s;
        src += s;
    }
}

// 15tap LPF(not-aligned)
NOA static void lpf_v_15_sse2u(pix * restrict dst, const pix * restrict src,
                               int h, int w, int s)
{
    int i,j;
    p_m128c src0, src1, src2, src3, src4, src5, src6, src7;
    p_m128c src8, src9, srca, srcb, srcc, srcd, srce;
    p_m128r dst0;
    __m128i xs0,xs1,xs2,xs3,xs4;

    w = COUNT16(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m128r)dst;
        src7 = (p_m128c)src;
        src6 = (p_m128c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src8 = (p_m128c)(src + s * ((i+1) >= h ? -1 : 1));
        src5 = (p_m128c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src9 = (p_m128c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src4 = (p_m128c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        srca = (p_m128c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src3 = (p_m128c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        srcb = (p_m128c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src2 = (p_m128c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srcc = (p_m128c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        src1 = (p_m128c)(src + s * ((i-6) >= 0 ? -6 : 6 - 2*i));
        srcd = (p_m128c)(src + s * ((i+6) >= h ? 2*(h-1-i)-6 : 6));
        src0 = (p_m128c)(src + s * ((i-7) >= 0 ? -7 : 7 - 2*i));
        srce = (p_m128c)(src + s * ((i+7) >= h ? 2*(h-1-i)-7 : 7));
        
        for(j=0; j<w; ++j){
            xs2 = _mm_adds_epi16(_mm_loadu_si128(src0+j), _mm_loadu_si128(srce+j));
            xs1 = _mm_adds_epi16(_mm_loadu_si128(src1+j), _mm_loadu_si128(srcd+j));
            xs0 = xs1;
            
            xs0 = _mm_unpackhi_epi16(xs0, xs2);
            xs1 = _mm_unpacklo_epi16(xs1, xs2);
            
            xs0 = _mm_madd_epi16(xs0, xc0_f.x);
            xs1 = _mm_madd_epi16(xs1, xc0_f.x);
            
            xs0 = _mm_add_epi32(xs0, xc4_f.x);
            xs1 = _mm_add_epi32(xs1, xc4_f.x);
            
            //
            xs4 = _mm_adds_epi16(_mm_loadu_si128(src2+j), _mm_loadu_si128(srcc+j));
            xs3 = _mm_adds_epi16(_mm_loadu_si128(src3+j), _mm_loadu_si128(srcb+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);

            xs2 = _mm_madd_epi16(xs2, xc1_f.x);
            xs3 = _mm_madd_epi16(xs3, xc1_f.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_adds_epi16(_mm_loadu_si128(src4+j), _mm_loadu_si128(srca+j));
            xs3 = _mm_adds_epi16(_mm_loadu_si128(src5+j), _mm_loadu_si128(src9+j));
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);

            xs2 = _mm_madd_epi16(xs2, xc2_f.x);
            xs3 = _mm_madd_epi16(xs3, xc2_f.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            //
            xs4 = _mm_adds_epi16(_mm_loadu_si128(src6+j), _mm_loadu_si128(src8+j));
            xs3 = _mm_loadu_si128(src7+j);
            xs2 = xs3;
            
            xs2 = _mm_unpackhi_epi16(xs2, xs4);
            xs3 = _mm_unpacklo_epi16(xs3, xs4);
            
            xs2 = _mm_madd_epi16(xs2, xc3_f.x);
            xs3 = _mm_madd_epi16(xs3, xc3_f.x);
            
            xs0 = _mm_add_epi32(xs0, xs2);
            xs1 = _mm_add_epi32(xs1, xs3);
            
            xs0 = _mm_srai_epi32(xs0, 15);
            xs1 = _mm_srai_epi32(xs1, 15);
            
            _mm_storeu_si128(dst0+j, _mm_packs_epi32(xs1, xs0));
        }
        
        dst += s;
        src += s;
    }
}


VLPF const vflt_sse2[]  = 
{ lpf_v_3_sse2,  lpf_v_5_sse2,  lpf_v_7_sse2,
  lpf_v_9_sse2, lpf_v_11_sse2, lpf_v_13_sse2, lpf_v_15_sse2, NULL };

VLPF const vflt_sse2u[]  = 
{ lpf_v_3_sse2u,  lpf_v_5_sse2u,  lpf_v_7_sse2u,
  lpf_v_9_sse2u, lpf_v_11_sse2u, lpf_v_13_sse2u, lpf_v_15_sse2u, NULL };

